#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <cglm/cglm.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "move.h"
#include "menus.h"
#include "proj.h"
#include "entities.h"
#include "generation.h"
#include "display.h"

int flashlight = 10 ;

int MAX_SIZE = 8192 ;

int* drawOrder ;

GLint loc_scale;
GLint loc_model;
GLint loc_view;
GLint loc_proj;

GLint loc_frag;

void init_draworder() {
    drawOrder = malloc(sizeof(int)*6) ;
}

float vertices_tr[9] ;
void gl_renderTriangle(unsigned int shaderProgram, unsigned int VAO, unsigned int VBO, double x0, double y0, double z0, double x1, double y1, double z1, double x2, double y2, double z2, float r, float g, float b) {
    vertices_tr[0] = (float)x0;
    vertices_tr[1] = (float)y0;
    vertices_tr[2] = (float)z0;
    vertices_tr[3] = (float)x1;
    vertices_tr[4] = (float)y1;
    vertices_tr[5] = (float)z1;
    vertices_tr[6] = (float)x2;
    vertices_tr[7] = (float)y2;
    vertices_tr[8] = (float)z2;

    glBindBuffer(GL_ARRAY_BUFFER, VBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertices_tr), vertices_tr, GL_STATIC_DRAW);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // note that this is allowed, the call to glVertexAttribPointer registered VBO as the vertex attribute's bound vertex buffer object so afterwards we can safely unbind
    glBindBuffer(GL_ARRAY_BUFFER, 0); 

    // You can unbind the VAO afterwards so other VAO calls won't accidentally modify this VAO, but this rarely happens. Modifying other
    // VAOs requires a call to glBindVertexArray anyways so we generally don't unbind VAOs (nor VBOs) when it's not directly necessary.
    glBindVertexArray(0); 

    glUseProgram(shaderProgram);
    glBindVertexArray(VAO);
    glDrawArrays(GL_TRIANGLES, 0, 3);
}

//double near = 0.4 ;
double near = 0.0 ;
double far = 10.0 ;

double top = 1.0 ;
double bottom = -1.0 ;
double left = -1.0 ;
double right = 1.0 ;

mat4 model, view, projection;
mat4 scale;
float vertices[108];

// -x ; +x ; -y ; +y ; -z ; +z //
void init_vertices() {
    vertices[0] = -0.5f; vertices[1] = -0.5f; vertices[2] = -0.5f;
    vertices[3] = -0.5f; vertices[4] =  0.5f; vertices[5] = -0.5f;
    vertices[6] = -0.5f; vertices[7] =  0.5f; vertices[8] =  0.5f;
    vertices[9] = -0.5f; vertices[10] = -0.5f; vertices[11] = -0.5f;
    vertices[12] = -0.5f; vertices[13] =  0.5f; vertices[14] =  0.5f;
    vertices[15] = -0.5f; vertices[16] = -0.5f; vertices[17] =  0.5f;

    vertices[18] =  0.5f; vertices[19] = -0.5f; vertices[20] = -0.5f;
    vertices[21] =  0.5f; vertices[22] =  0.5f; vertices[23] =  0.5f;
    vertices[24] =  0.5f; vertices[25] =  0.5f; vertices[26] = -0.5f;
    vertices[27] =  0.5f; vertices[28] = -0.5f; vertices[29] = -0.5f;
    vertices[30] =  0.5f; vertices[31] = -0.5f; vertices[32] =  0.5f;
    vertices[33] =  0.5f; vertices[34] =  0.5f; vertices[35] =  0.5f;

    vertices[36] = -0.5f; vertices[37] = -0.5f; vertices[38] = -0.5f;
    vertices[39] =  0.5f; vertices[40] = -0.5f; vertices[41] =  0.5f;
    vertices[42] =  0.5f; vertices[43] = -0.5f; vertices[44] = -0.5f;
    vertices[45] = -0.5f; vertices[46] = -0.5f; vertices[47] = -0.5f;
    vertices[48] = -0.5f; vertices[49] = -0.5f; vertices[50] =  0.5f;
    vertices[51] =  0.5f; vertices[52] = -0.5f; vertices[53] =  0.5f;

    vertices[54] = -0.5f; vertices[55] =  0.5f; vertices[56] = -0.5f;
    vertices[57] =  0.5f; vertices[58] =  0.5f; vertices[59] = -0.5f;
    vertices[60] =  0.5f; vertices[61] =  0.5f; vertices[62] =  0.5f;
    vertices[63] = -0.5f; vertices[64] =  0.5f; vertices[65] = -0.5f;
    vertices[66] =  0.5f; vertices[67] =  0.5f; vertices[68] =  0.5f;
    vertices[69] = -0.5f; vertices[70] =  0.5f; vertices[71] =  0.5f;

    vertices[72] = -0.5f; vertices[73] = -0.5f; vertices[74] = -0.5f;
    vertices[75] =  0.5f; vertices[76] = -0.5f; vertices[77] = -0.5f;
    vertices[78] =  0.5f; vertices[79] =  0.5f; vertices[80] = -0.5f;
    vertices[81] = -0.5f; vertices[82] = -0.5f; vertices[83] = -0.5f;
    vertices[84] =  0.5f; vertices[85] =  0.5f; vertices[86] = -0.5f;
    vertices[87] = -0.5f; vertices[88] =  0.5f; vertices[89] = -0.5f;

    vertices[90] = -0.5f; vertices[91] = -0.5f; vertices[92] =  0.5f;
    vertices[93] =  0.5f; vertices[94] =  0.5f; vertices[95] =  0.5f;
    vertices[96] =  0.5f; vertices[97] = -0.5f; vertices[98] =  0.5f;
    vertices[99] = -0.5f; vertices[100] = -0.5f; vertices[101] =  0.5f;
    vertices[102] = -0.5f; vertices[103] =  0.5f; vertices[104] =  0.5f;
    vertices[105] =  0.5f; vertices[106] =  0.5f; vertices[107] =  0.5f;
}

void gl_renderCube(unsigned int shaderProgram, unsigned int fragmentShader, unsigned int VAO, unsigned int VBO, cube_0* c, double offx, double offy, double offz) {
    glm_mat4_identity(model);
    glm_mat4_identity(scale);
    scale[0][0] = (float)(c->w);
    scale[1][1] = (float)(c->h);
    scale[2][2] = (float)(c->d);
    glm_translate(model, (vec3){(float)(c->x+c->w/2.0+offx), (float)(c->y+c->h/2.0+offy), (float)(c->z+c->d/2.0+offz)});
    glm_rotate(model, (float)(-c->hz_angle), (vec3){0.0f, 1.0f, 0.0f});
    glm_rotate(model, (float)(c->vt_angle), (vec3){1.0f, 0.0f, 0.0f});

    glUniformMatrix4fv(loc_scale, 1, GL_FALSE, (float*)scale);
    glUniformMatrix4fv(loc_model, 1, GL_FALSE, (float*)model);
    glUniformMatrix4fv(loc_view, 1, GL_FALSE, (float*)view);
    glUniformMatrix4fv(loc_proj, 1, GL_FALSE, (float*)projection);

    glUniform4f(loc_frag, c->red/255.0f, c->green/255.0f, c->blue/255.0f, 1.0f);

    glDrawArrays(GL_TRIANGLES, 0, 36);
    triCount += 12;
}

double px, py, pz;
double px2, py2, pz2;
bool is_visible(cube_0* cb, double offx, double offy, double offz) {
    for(int d = 0; d < 8; d++) {
        project_to_cube(cb->x+cb->w*(d%2==0)+offx, cb->y+cb->h*((d/2)%2==0)+offy, cb->z+cb->d*((d/4)%2==0)+offz, &px, &py, &pz, cb);
        project_to_camera(px, py, pz, &px2, &py2, &pz2);
        if(pz2 >= near) {
            return true;
        }
    }
    return false;
}

void gl_renderAll(unsigned int shaderProgram, unsigned int fragmentShader, unsigned int VAO, unsigned int VBO, room* rtd, double offx, double offy, double offz) {
    if(rtd != NULL) {
        for(int k = 0; k < rtd->map_size; k++) {
            if(is_visible(rtd->map[k], offx, offy, offz)) {
                gl_renderCube(shaderProgram, fragmentShader, VAO, VBO, rtd->map[k], offx, offy, offz);
            }
        }
        for(int k = 0; k < rtd->tps_size; k++) {
            if(is_visible(rtd->tps[k]->hitbox, offx, offy, offz)) {
                gl_renderCube(shaderProgram, fragmentShader, VAO, VBO, rtd->tps[k]->hitbox, offx, offy, offz);
            }
        }
        for(int k = 0; k < rtd->ent_len; k++) {
            if(is_visible(rtd->ents[k]->pos, offx, offy, offz)) {
                gl_renderCube(shaderProgram, fragmentShader, VAO, VBO, rtd->ents[k]->pos, offx, offy, offz);
            }
        }
    }
}

void gl_renderNearbyChunks(unsigned int shaderProgram, unsigned int fragmentShader, unsigned int VAO, unsigned int VBO, int render_distance) {
    for(int w = -render_distance; w <= render_distance; w++) {
        for(int h = -render_distance; h <= render_distance; h++) {
            gl_renderAll(shaderProgram, fragmentShader, VAO, VBO, hashtbl_find_opt(visited, player_chx+w, player_chy+h), (2.0*room_width)*w, 0.0, (2.0*room_depth)*h);
        }
    }
}

vec3 dir0;
vec3 direction;
void gl_initRender(unsigned int shaderProgram, unsigned int fragmentShader, unsigned int VAO, unsigned int VBO) {
    dir0[0] = sinf((float)(rot_hz)) * cosf((float)(rot_vt));
    dir0[1] = -sinf((float)(rot_vt));
    dir0[2] = cosf((float)(rot_hz)) * cosf((float)(rot_vt));

    glm_vec3_normalize(dir0); // Normalize to unit length
    glm_vec3_add((vec3){(float)camx, (float)camy, (float)camz}, dir0, direction);

    glm_mat4_identity(view);
    glm_lookat((vec3){(float)camx, (float)camy, (float)camz}, direction, (vec3){0.0f, 1.0f, 0.0f}, view);

    glm_perspective(glm_rad((float)fov), 1500.0f / 1000.0f, 0.1f, 100.0f, projection);

    loc_scale = glGetUniformLocation(shaderProgram, "scale");
    loc_model = glGetUniformLocation(shaderProgram, "model");
    loc_view = glGetUniformLocation(shaderProgram, "view");
    loc_proj = glGetUniformLocation(shaderProgram, "projection");

    loc_frag = glGetUniformLocation(fragmentShader, "u_color");
}

void gl_drawData(unsigned int shaderProg) {
    gl_drawInteger(shaderProg, (int)camx, 0.95f, 0.9f, 0.05f, 255, 255, 255, 0.005, -1);
    gl_drawInteger(shaderProg, (int)camy, 0.95f, 0.75f, 0.05f, 255, 255, 255, 0.005, -1);
    gl_drawInteger(shaderProg, (int)camz, 0.95f, 0.6f, 0.05f, 255, 255, 255, 0.005, -1);

    gl_drawInteger(shaderProg, (int)10.0*camvx, 0.95f, -0.6f, 0.05f, 255, 255, 255, 0.005, -1);
    gl_drawInteger(shaderProg, (int)10.0*camvy, 0.95f, -0.75f, 0.05f, 255, 255, 255, 0.005, -1);
    gl_drawInteger(shaderProg, (int)10.0*camvz, 0.95f, -0.9f, 0.05f, 255, 255, 255, 0.005, -1);

    gl_drawInteger(shaderProg, player_hp, -0.95f, 0.9f, 0.05f, 255-player_hp/4, player_hp/4, 0, 0.005f, 1);
}